#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

void perror(const char *s);

#include <errno.h>

#include "../include/m4v.h"

unsigned char prev_frame[256][64], cur_frame[256][64];
unsigned char screen[256][128];

/*
	The diffmap holds the number of blocks which differ per line and their addresses, hence 65 entries per line.
*/

int diffmap[256][65];

#define QLSCRSIZ 256*128

int encode_video(int numfiles, char *filelist[])
{
	int i, j, k, lastchange;
	int fd;
	unsigned char codebyte;

/* First of all, before we start encoding clear out the previous frame to a known state. */

	for (i = 0; i < 256; i++)
		for (j = 0; j < 64; j++)
			prev_frame[i][j] = (unsigned char) 0x00;

/*
	For each of the files in the list process them.
*/

	for (i = 0; i < numfiles; i++)
	{
/*
		Open the files and read them in.
*/
#ifdef DEBUG
		fprintf(stderr, "Opening %s: file %d of %d\n", filelist[i], i + 1, numfiles);
#endif
		if ((fd = open(filelist[i], O_RDONLY)) < 0)
		{
			perror(filelist[i]);
			exit;
		}

		if (read(fd, screen, QLSCRSIZ) != QLSCRSIZ)
		{
			fprintf(stderr, "QL Screen image too small: ");
			perror(filelist[i]);
		}

		close(fd);

/*
		Zero out the frame map changed block counts.
*/

		for (j = 0; j < 256; diffmap[j++][0] = 0);

/*
		Do a quick and dirty copy assuming that it's a monochrome
		bitmap and map out which pixel blacks have changed since the
		previous frame.
*/

		for (j = 0; j < 256; j++)
		{
			for (k = 0; k < 64; k++)
			{
				cur_frame[j][k] = screen[j][k*2];


/*

				If the block in the current frame differs from
				the one in the previous frame then increment
				the map's counter and write the current block
				address into the next locatinnn n the list.

*/
#ifdef DEBUG
				fprintf(stderr, "cur_frame[%d][%d] = %02x prev_frame[%d][%d] = %02x\n",
					j, k, cur_frame[j][k],
					j, k, prev_frame[j][k]);
#endif

				if (cur_frame[j][k] != prev_frame[j][k])
				{
					diffmap[j][0]++;
					diffmap[j][diffmap[j][0]] = k;
#ifdef DEBUG
					fprintf(stderr, "Difference found: diffmap[%d] is now %d for block %d.\n",
						j, diffmap[j][0], diffmap[j][diffmap[j][0]]);
#endif
				}
			}
		}

/*
		Now we find the last scan line which has any changes.
*/

		lastchange = 0;
		
		for (j = 255; j > 0 && lastchange == 0; j--)
			if (diffmap[j][0] > 0)
				lastchange = j;

#ifdef DEBUG
		fprintf(stderr, "Last changed scan line is %d\n", lastchange);
#endif

/*
		OK, we have all the information we need, the number of changes on each line
		and the last line which has any changes on it.

		We can now write out the data for the frame itself.

		Note: We do this one byte at a time so as to make absolutely sure that we don't have any endian problems.
*/

		for (j = 0; j <= lastchange; j++)
		{
			if (j == 0)
			{
				codebyte = START_FRAME;
#ifdef DEBUG
				fprintf(stderr, "******************* NEW BALLS PLEASE! ************************\n");
#endif
			}
			else
				codebyte = 0;
				
/* Test to see if the line has more than 50% change, if not then it's an difference line. */

			if (diffmap[j][0] < 32)
			{

/* The codebyte has the "difference" bit set and the last 6 bits set to being the number of changes. */
#ifdef DEBUG
				fprintf(stderr, "We have a difference line with %d changes.\n", diffmap[j][0]);
#endif

				codebyte = codebyte + DIFFERENCE_LINE + diffmap[j][0];

				write(1, &codebyte, sizeof(unsigned char));

				if (diffmap[j][0] > 0)
				{
#ifdef DEBUG
					fprintf(stderr, "Differences: ");
#endif
					for (k = 1; k <= diffmap[j][0]; k++)
					{
#ifdef DEBUG
						fprintf(stderr, "(%d,%02x) ", diffmap[j][k], cur_frame[j][diffmap[j][k]]);
#endif
						write(1, &diffmap[j][k], sizeof(unsigned char));
						write(1, &cur_frame[j][diffmap[j][k]], 1);
					}
#ifdef DEBUG
					fprintf(stderr, "\n");
#endif
				}
			}
			else
			{
#ifdef DEBUG
				fprintf(stderr, "We have a full line.\n");
#endif
				write(1, &codebyte, sizeof(unsigned char));
				
				for (k = 0; k < 64; k++)
					write(1, &cur_frame[j][k], 1);
			}
		}

/* Copy the data from the cur_frame into the prev_frame ready for the next iteration. */
		for (j = 0; j < 256; j++)
			for (k = 0; k < 64; k++)
				prev_frame[j][k] = cur_frame[j][k];
	}
}

#ifdef DEBUG

static char *testfiles[] = {
"test-scr/test-0.scr",
"test-scr/test-1.scr",
"test-scr/test-2.scr",
"test-scr/test-3.scr",
"test-scr/test-4.scr",
"test-scr/test-5.scr",
"test-scr/test-6.scr",
"test-scr/test-7.scr",
"test-scr/test-8.scr",
"test-scr/test-9.scr",
"test-scr/test-10.scr",
"test-scr/test-11.scr",
"test-scr/test-12.scr",
"test-scr/test-13.scr",
"test-scr/test-14.scr",
"test-scr/test-15.scr",
"test-scr/test-16.scr",
"test-scr/test-17.scr",
"test-scr/test-18.scr",
"test-scr/test-19.scr",
"test-scr/test-20.scr",
"test-scr/test-21.scr",
"test-scr/test-22.scr",
"test-scr/test-23.scr",
"test-scr/test-24.scr",
"test-scr/test-25.scr",
"test-scr/test-26.scr",
"test-scr/test-27.scr",
"test-scr/test-28.scr",
"test-scr/test-29.scr",
"test-scr/test-30.scr",
"test-scr/test-31.scr",
"test-scr/test-32.scr",
"test-scr/test-33.scr",
"test-scr/test-34.scr",
"test-scr/test-35.scr",
"test-scr/test-36.scr",
"test-scr/test-37.scr",
"test-scr/test-38.scr",
"test-scr/test-39.scr",
"test-scr/test-40.scr",
"test-scr/test-41.scr",
"test-scr/test-42.scr",
"test-scr/test-43.scr",
"test-scr/test-44.scr",
"test-scr/test-45.scr",
"test-scr/test-46.scr",
"test-scr/test-47.scr",
"test-scr/test-48.scr",
"test-scr/test-49.scr",
"test-scr/test-50.scr",
"test-scr/test-51.scr",
"test-scr/test-52.scr",
"test-scr/test-53.scr",
"test-scr/test-54.scr",
"test-scr/test-55.scr",
"test-scr/test-56.scr",
"test-scr/test-57.scr",
"test-scr/test-58.scr",
"test-scr/test-59.scr",
"test-scr/test-60.scr",
"test-scr/test-61.scr",
"test-scr/test-62.scr",
"test-scr/test-63.scr",
"test-scr/test-64.scr",
"test-scr/test-65.scr",
"test-scr/test-66.scr",
"test-scr/test-67.scr",
"test-scr/test-68.scr",
"test-scr/test-69.scr",
"test-scr/test-70.scr",
"test-scr/test-71.scr",
"test-scr/test-72.scr",
"test-scr/test-73.scr",
"test-scr/test-74.scr",
"test-scr/test-75.scr",
"test-scr/test-76.scr",
"test-scr/test-77.scr",
"test-scr/test-78.scr",
"test-scr/test-79.scr",
"test-scr/test-80.scr",
"test-scr/test-81.scr",
"test-scr/test-82.scr",
"test-scr/test-83.scr",
"test-scr/test-84.scr",
"test-scr/test-85.scr",
"test-scr/test-86.scr",
"test-scr/test-87.scr",
"test-scr/test-88.scr",
"test-scr/test-89.scr",
"test-scr/test-90.scr",
"test-scr/test-91.scr",
"test-scr/test-92.scr",
"test-scr/test-93.scr",
"test-scr/test-94.scr",
"test-scr/test-95.scr",
"test-scr/test-96.scr",
"test-scr/test-97.scr",
"test-scr/test-98.scr",
"test-scr/test-99.scr",
"test-scr/test-100.scr",
"test-scr/test-101.scr",
"test-scr/test-102.scr",
"test-scr/test-103.scr",
"test-scr/test-104.scr",
"test-scr/test-105.scr",
"test-scr/test-106.scr",
"test-scr/test-107.scr",
"test-scr/test-108.scr",
"test-scr/test-109.scr",
"test-scr/test-110.scr",
"test-scr/test-111.scr",
"test-scr/test-112.scr",
"test-scr/test-113.scr",
"test-scr/test-114.scr",
"test-scr/test-115.scr",
"test-scr/test-116.scr",
"test-scr/test-117.scr",
"test-scr/test-118.scr",
"test-scr/test-119.scr",
"test-scr/test-120.scr",
"test-scr/test-121.scr",
"test-scr/test-122.scr",
"test-scr/test-123.scr",
"test-scr/test-124.scr",
"test-scr/test-125.scr",
"test-scr/test-126.scr",
"test-scr/test-127.scr",
"test-scr/test-128.scr",
"test-scr/test-129.scr",
"test-scr/test-130.scr",
"test-scr/test-131.scr",
"test-scr/test-132.scr",
"test-scr/test-133.scr",
"test-scr/test-134.scr",
"test-scr/test-135.scr",
"test-scr/test-136.scr",
"test-scr/test-137.scr",
"test-scr/test-138.scr",
"test-scr/test-139.scr",
"test-scr/test-140.scr",
"test-scr/test-141.scr",
"test-scr/test-142.scr",
"test-scr/test-143.scr",
"test-scr/test-144.scr",
"test-scr/test-145.scr",
"test-scr/test-146.scr",
"test-scr/test-147.scr",
"test-scr/test-148.scr",
"test-scr/test-149.scr"
};

main()
{
	encode_video(150, testfiles);
}

#endif
